/******************************************************
*
*	Copyright by Kamil Szatkowski "Netrix"
*
*	This code is distributed as Public Domain
*
******************************************************/
#include "GraphColoring.hpp"
#include <stdio.h>
#include <math.h>
#include <string.h>

#include <algorithm>
using std::min_element;

#ifdef _WIN32
	#include <windows.h>
#endif

void drawGraph(GraphColoring& graph, FILE* pFile = 0);
uint SelectIndividual(const uint* pGenerationConflicts, uint uGenerationSize, uint uExcept);
void CrossoverIndividuals(const uint* pFirstInd, const uint* pSecondInd, uint uIndividualSize, uint* pNewFirstInd, uint* pNewSecondInd);
void MutateIndividual(uint* pIndividual, uint uIndividualSize, uint uColorsCount);

int main()
{
	GraphColoring graph(5000, 0.4, 0.2, SelectIndividual, CrossoverIndividuals, MutateIndividual, 9);

	if(graph.LoadGraph("queen8_8.col"))
	{
		printf("Failed to load graph: %s", "queen8_8.col");
		return 0;
	}

	graph.InitPopulation();

	uint i;
	ull uSecond, uFirst;// = __rdtsc();

#ifdef _WIN32
	ull uFreq;
	QueryPerformanceFrequency((LARGE_INTEGER*)&uFreq);

	QueryPerformanceCounter((LARGE_INTEGER*)&uFirst);
#else	// GCC
	uFirst = __rdtsc();
#endif

	for(i = 0; i < 1000/* && graph.GetBestConflictsCount() != 0*/; ++i)	// Commented out for time tests
	{
		graph.IterateGeneration();
	}
#ifdef _WIN32
	QueryPerformanceCounter((LARGE_INTEGER*)&uSecond);
#else
	uSecond = __rdtsc();
#endif
	printf("stop\n");

	FILE* pFile = fopen("Tests_out.txt", "a+");
#ifdef _WIN32
	fprintf(pFile, "Clock frequency: %Lu, ", uFreq);
#endif
	fprintf(pFile, "It: %d, G: %d, I: %d, c: %d, delta-cycles: %Lu\n", i, graph.GetBestGeneration(), graph.GetBestIndividual(), graph.GetBestConflictsCount(), uSecond - uFirst);
	fclose(pFile);

	drawGraph(graph);
	graph.ReleaseGraph();

	return 0;
}

///////////////////////////////////////////////////////
class TournamentComparator
{
public:
	TournamentComparator(const uint* pGenConflicts, uint uExcept)
		: m_pGenConflicts(pGenConflicts)
		, m_uExcept(uExcept)
	{}
	
	bool operator()(uint a, uint b) const
	{
		return m_pGenConflicts[a] < m_pGenConflicts[b] && m_pGenConflicts[a] != m_uExcept;
	}

private:
	const uint* m_pGenConflicts;
	uint m_uExcept;
};

template<uint RANGE>
class RandomSet
{
public:
	RandomSet()
	{ 
		for(uint i = 0; i < RANGE; ++i)
		{ m_aSet[i] = i; }
	}

	const uint* GetRandomSet(uint uCount, uint uInnerRange)
	{
		uint uTemp, uRand;
		for(uint i = 0; i < uCount; ++i)
		{
			uRand = rand() % (uInnerRange - i);
			uTemp = m_aSet[i];
			m_aSet[i] = m_aSet[i + uRand];
			m_aSet[i + uRand] = uTemp;
		}
		return m_aSet;
	}
	
private:
	uint m_aSet[RANGE];
};

RandomSet<5000> g_Tourn;	// min
///////////////////////////////////////////////////////

uint SelectIndividual(const uint* pGenerationConflicts, uint uGenerationSize, uint uExcept)
{
	const double TOURNAMENT_SIZE_COEFF = 0.01;

	// Tournament
	uint uTournamentSize = (uint)(uGenerationSize * TOURNAMENT_SIZE_COEFF);	// 20 bylo w miare
	const uint* pSet = g_Tourn.GetRandomSet(uTournamentSize, uGenerationSize);

	TournamentComparator cmp(pGenerationConflicts, uExcept);

	return (*min_element(pSet, pSet + uTournamentSize, cmp));
}

void CrossoverIndividuals(const uint* pFirstInd, const uint* pSecondInd, uint uIndividualSize, uint* pNewFirstInd, uint* pNewSecondInd)
{
	uint uCrossingPoint = rand() % uIndividualSize;

	// First part
	uint uWall = uCrossingPoint * sizeof(uint);
	memcpy(pNewFirstInd, pFirstInd, uWall);
	memcpy(pNewSecondInd, pSecondInd, uWall);

	// Second part
	uint uBreak = uIndividualSize * sizeof(uint) - uWall;
	memcpy(pNewFirstInd, pSecondInd + uCrossingPoint, uBreak);
	memcpy(pNewSecondInd, pFirstInd + uCrossingPoint, uBreak);
}

void MutateIndividual(uint* pIndividual, uint uIndividualSize, uint uColorsCount)
{
	const double MUTATE_RANGE_COEFF = 0.1;

	uint uMutateRange = (uint)(uIndividualSize * MUTATE_RANGE_COEFF);
	// Oh, my lucky Mutation
	uint uMP = rand() % uIndividualSize;
	
	for(uint i = 0; i < uMutateRange; ++i)
	{
		if(uMP == uMutateRange)
		{
			uMP = 0;
		}

		pIndividual[uMP] = rand() % uColorsCount;
	}
}

void drawGraph(GraphColoring& graph, FILE* pFile)
{
	uint fields = graph.GetFieldsCount();
	uint uSqrt = (uint)sqrtf((float)fields);

	if(pFile != 0)
	{
		fprintf(pFile, "Graf k = %d, conflicts = %d:\n", graph.GetColorsCount(), graph.GetBestConflictsCount());
	}
	else
	{
		printf("Graf k = %d, conflicts = %d:\n", graph.GetColorsCount(), graph.GetBestConflictsCount());
	}

	for(uint uField = 0; uField < fields; )
	{
		for(uint i = 0; uField < fields && i < uSqrt; ++i, ++uField)
		{
			if(pFile != 0)
			{
				fprintf(pFile, "%3d ", graph.GetBestFieldColor(uField));
			}
			else
			{
				printf("%3d ", graph.GetBestFieldColor(uField));
			}
		}
		pFile != 0 ? fprintf(pFile, "\n") : printf("\n");
	}

	pFile != 0 ? fprintf(pFile, "\n") : printf("\n");
}